/*******************************************************************

  splinevect.c : Uses splining to redistribute vertices on a plCurve
                 to obtain edges of given edgelength or according to 
                 other criteria.

	   ************************************************************/

#include<splinevect.h>
#include<argtable2.h>

extern int  octrope_error_num;
extern char octrope_error_str[80];

/**********************************************************************/

struct arg_lit  *verbose;
struct arg_file *corefile;
struct arg_lit  *help;
struct arg_dbl  *edgelength;
struct arg_dbl  *resolution;
struct arg_lit  *mps;
struct arg_end  *end;
struct arg_end  *helpend;

plCurve *core;
FILE    *infile_fptr,*outfile_fptr;

bool resolution_ok(plCurve *inLink,int *nv) 

{
  bool ok=true;
  int i;

  for(i=0;i<inLink->nc;i++) {

    if (inLink->cp[i].nv < nv[i]) { ok = false; }
    
  }

  return ok;

}


/****************************** Main procedure ********************************/
  
int main(int argc,char *argv[])
{
  int            infilenum,j,nerrors;
  char           outfile_tail[1000];
 
  void *argtable[] = 
    {
     verbose = arg_lit0("v","verbose","print debugging information"),
     corefile  = arg_filen(NULL,NULL,"<file>",1,100000,"input files"),
     edgelength = arg_dbl0("e","edgelength","<x>","splines to edges of length < x"),
     resolution = arg_dbl0("r","resolution","<n>","spline to n vertices per unit length"),
     mps = arg_lit0(NULL,"mps","minrad preserving spline (instead of cubic)"),
     help = arg_lit0(NULL,"help","display help message"),
     end = arg_end(20)};
  
  void *helptable[] = {help,helpend = arg_end(20)};
  
  /* First, we parse the command-line arguments using argtable. */

  if (arg_nullcheck(argtable) != 0)
    printf("splinevect: Insufficient memory to allocate argument table.\n");

  nerrors = arg_parse(argc,argv,argtable);

  if (nerrors > 0) { /* The standard syntax didn't match-- try the help table */
    
    nerrors = arg_parse(argc,argv,helptable);
    
    if (nerrors > 0) {  /* The help table didn't match either-- the
                         first set of errors was probably more
                         helpful, so we display it. */

      arg_print_errors(stdout,end,"splinevect");
      exit(1);

    } else {  /* The help table matched, which means we 
		 asked for help or gave nothing */
  
      printf("splinevect redistributes vertices on a VECT polyline\n"
           "usage: \n\n");
      arg_print_glossary(stdout, argtable," %-25s %s\n");
      exit(0);

    }
    
  }

  if (edgelength->count == 0 && resolution->count == 0) {
  
    printf("splinevect redistributes vertices on a VECT polyline\n"
	   "usage: \n\n");
    arg_print_glossary(stdout, argtable," %-25s %s\n");
    exit(0);

  }
    
  double res = 0.001,len = 1000;

  if (resolution->count > 0) {

    res = resolution->dval[0];

  }

  if (edgelength->count > 0) {

    len = edgelength->dval[0];

  }

  res = (res > 1.0/len) ? res : 1.0/len;

  /* Now we have parsed the arguments and are ready to work. */
  
  for(infilenum = 0;infilenum < corefile->count;infilenum++) {

    /* We begin by loading the current core link. */

     infile_fptr = fopen(corefile->filename[infilenum],"r");
  
     if (infile_fptr == NULL) {
    
       fprintf(stderr,"splinevect: Couldn't open file %s.\n",corefile->filename[infilenum]);
       continue;  /* Try the next file */
 
     }
  
     int plr_error_num;
     char plr_error_str[1024];

     core = plc_read(infile_fptr,
		     &plr_error_num,plr_error_str,sizeof(plr_error_str));
  
     if (plr_error_num > 0) {   /* This is the signal for an error. */
    
       fprintf(stderr,"splinevect: link reading error\n%s\n",plr_error_str);
       continue;  /* Try the next file */
    
     }
  
     fclose(infile_fptr);

     printf("splinevect: Read link from %s.\n",
	    corefile->basename[infilenum]);
     
     /* Now we actually do the splining... */

     plCurve *newCurve = NULL;
     plc_spline *spline = NULL;
     bool ok;
     double totlen;

     double     *length = malloc(sizeof(double)*core->nc);
     int        *nv = malloc(sizeof(int)*core->nc);
     int        *ne = malloc(sizeof(int)*core->nc);
     int        *nne = malloc(sizeof(int)*core->nc);

     assert(nv != NULL);
     assert(length != NULL);
     assert(ne != NULL);
     assert(nne != NULL);

     totlen = plc_arclength(core,length);
     for(j=0;j<core->nc;j++) {nv[j] = ceil(res*length[j]);}   

     if (mps->count == 0) {

       spline = plc_convert_to_spline(core,&ok);
       assert(ok);
       newCurve = plc_convert_from_spline(spline,nv);

     } else { 

       int failsafe;
       plCurve *swapCurve;

       newCurve = plc_copy(core);

       for(failsafe=0;!resolution_ok(newCurve,nv);failsafe++) {

	 swapCurve = plc_double_verts(newCurve);
	 plc_free(newCurve);
	 newCurve = plc_copy(swapCurve);
	 plc_free(swapCurve);

	 assert(failsafe < 10);

       }

     }
       
     printf("splinevect: Converted %g unit, %d edge plCurve ( ",totlen,plc_edges(core,ne));
     for(j=0;j<core->nc;j++) { printf("%d ",ne[j]); } 
     printf(")\n");

     printf("            To %d edge plCurve ( ",plc_edges(newCurve,nne));
     for(j=0;j<core->nc;j++) { printf("%d ",nne[j]); } 
     printf(")\n");

     char outfile_name[16000];
     
     sprintf(outfile_name,"%s",corefile->basename[infilenum]);
     
     if (strstr(outfile_name,".vect") != NULL) {
       
       sprintf(strstr(outfile_name,".vect"),".r%d.vect",(int)(ceil(res)));
       
     } else {
       
       sprintf(outfile_tail,".r%d.vect",(int)(ceil(res)));	 
       strcat(outfile_name,outfile_tail);
       
     }
     
     outfile_fptr = fopen(outfile_name,"w");
       
     if (outfile_fptr == NULL) {
       
       fprintf(stderr,
	       "splinevect: Couldn't open %s for writing.\n",outfile_name);
       exit(1);
	 
     }
       
     /* Now write the component to the file. */
       
     plc_write(outfile_fptr,newCurve);
     fclose(outfile_fptr);

     printf("splinevect: Wrote link to %s.\n",
	    outfile_name);

     plc_free(newCurve);
     plc_spline_free(spline);
     free(nv);
     free(length);
     free(ne);
     free(nne);
     plc_free(core);    
     
  }

  if (verbose->count > 0) {

    printf("\nsplinevect processed %d input files.\n",infilenum);

  }

  return 0;

}




